package be.iminds.ilabt.jfed.rspec.model.imutable_impl;

import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

public class ImmutableRspecNode implements RspecNode {
    private final static Logger LOG = LoggerFactory.getLogger(ImmutableRspecNode.class);

    private final ImmutableModelRspec modelRspec;

    protected final List<SliverType> sliverTypes;
    final List<ImmutableRspecInterface> interfaces;
    private final String clientId;
    private final double editorPosX;
    private final double editorPosY;
    private final double flackEditorPosX;
    private final double flackEditorPosY;
    private final List<HardwareType> hardwareTypes;
    private final Boolean exclusive;
    private final Boolean available;
    private final NodeLocation nodeLocation;
    private final GeniUrn componentManagerId;
    private final GeniUrn componentId;
    private final String componentName;
    private final GeniUrn sliverId;
    private final String hostname;
    private final boolean hasRoutableControlIp;
    private final List<String> leaseIdRefs;

    private final List<ImmutableExecuteService> executeServices;
    private final List<ImmutableInstallService> installServices;
    private final List<LoginService> loginServices;
    private final List<ProxyService> proxyServices;
    private final List<WebVncService> webVncServices;
    private final List<ImmutableAnsibleService> ansibleServices;
    private final List<String> ansibleGroups;
    private final List<ExtraXml> extraServicesXmls;

    private final List<ExtraXml> extraXmls;
    private final Set<EmptyTag> emptyTags;

    @Override
    public ImmutableModelRspec getRspec() {
        return modelRspec;
    }

    private static <T> List<T> listInit(List<T> orig) {
        if (orig == null) return null;
        return Collections.unmodifiableList(new ArrayList<>(orig));
    }
    private static <T> Set<T> setInit(Collection<T> orig) {
        if (orig == null) return null;
        return Collections.unmodifiableSet(new TreeSet<>(orig));
    }
    private static <OT, NT> List<NT> listInit(List<OT> orig, Function<OT, NT> convert) {
        if (orig == null) return null;
        return Collections.unmodifiableList(orig.stream().map(convert).collect(Collectors.toList()));
    }
    
    ImmutableRspecNode(ImmutableModelRspec newModelRspec, RspecNode orig) {
        this.modelRspec = newModelRspec;
        this.clientId = orig.getClientId();
        this.editorPosX = orig.getEditorX();
        this.editorPosY = orig.getEditorY();
        this.flackEditorPosX = orig.getFlackEditorX();
        this.flackEditorPosY = orig.getFlackEditorY();
        this.componentId = orig.getComponentId();
        this.componentManagerId = orig.getComponentManagerId();
        this.componentName = orig.getComponentName();
        this.sliverId = orig.getSliverId();
        this.exclusive = orig.getExclusive();
        this.available = orig.getAvailable();
        this.hostname = orig.getHostName();
        this.nodeLocation = orig.getLocation();
        this.leaseIdRefs = listInit(orig.getLeaseIdRefs());

        this.executeServices = listInit(orig.getExecuteServices(), ImmutableExecuteService::new);
        this.hardwareTypes = listInit(orig.getHardwareTypes());
        this.installServices = listInit(orig.getInstallServices(), ImmutableInstallService::new);
        this.loginServices = listInit(orig.getLoginServices());
        this.proxyServices = listInit(orig.getProxyServices());
        this.webVncServices = listInit(orig.getWebVncServices());
        this.ansibleServices = listInit(orig.getAnsibleServices(), ImmutableAnsibleService::new);
        this.ansibleGroups = listInit(orig.getAnsibleGroups());
        this.sliverTypes = listInit(orig.getSliverTypes());
        this.hasRoutableControlIp = orig.getRoutableControlIp();

        this.emptyTags = setInit(orig.getEmptyTags());
        this.extraXmls = listInit(orig.getExtraXml());
        this.extraServicesXmls = listInit(orig.getExtraServicesXml());

        List<ImmutableRspecInterface> copiedInterfaces = new ArrayList<>();
        for (RspecInterface origInterface : orig.getInterfaces()) {
            ImmutableRspecInterface copiedIface = newModelRspec.getInterfaceByClientId(origInterface.getClientId());
            if (copiedIface != null) {
                //use already copied iface, but bind to it.
                copiedIface.bindNode(this);
            } else {
                //clone iface
                copiedIface = new ImmutableRspecInterface(this.getRspec(),this, null, origInterface);
            }
            copiedInterfaces.add(copiedIface);
        }
        this.interfaces = Collections.unmodifiableList(copiedInterfaces);
    }

    @Override
    public String getClientId() {
        return clientId;
    }

    @Override
    public void setClientId(String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    @Nonnull
    public String getUniqueId() {
        return BasicStringRspec.makeUniqueNodeId(
                clientId,
                componentId != null ? componentId.getValue() : null,
                componentManagerId != null ? componentManagerId.getValue() : null
        );
    }

    @Override
    public double getEditorX() {
        return editorPosX;
    }

    @Override
    public void setEditorX(double value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public double getEditorY() {
        return editorPosY;
    }

    @Override
    public void setEditorY(double value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public double getFlackEditorX() {
        return flackEditorPosX;
    }

    @Override
    public void setFlackEditorX(double value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public double getFlackEditorY() {
        return flackEditorPosY;
    }

    @Override
    public void setFlackEditorY(double value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<HardwareType> getHardwareTypes() {
        return hardwareTypes;
    }

    @Override
    public Boolean getExclusive() {
        return exclusive;
    }

    @Override
    public void setExclusive(Boolean value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public Boolean getAvailable() {
        return available;
    }

    @Override
    public void setAvailable(Boolean available) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setSliverTypeDiskImage(String diskImageName) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getSliverTypeName() {
        if (sliverTypes.size() > 1)
            throw new RuntimeException("The caller did not expect this RSpec to have multiple sliver types. (it probably expected this to be a request rspec, which does indeed not have multiple sliver_type's)");

        if (sliverTypes.isEmpty()) return null;
        return sliverTypes.get(0).getName();
    }

    @Override
    public void setSliverTypeName(String name) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SliverType getSliverType() {
        if (sliverTypes.isEmpty())
            return null;
        if (sliverTypes.size() > 1)
            throw new RuntimeException("The caller did not expect this RSpec to have multiple sliver types. (it probably expected this to be a request rspec, which does indeed not have multiple sliver_type's)");
        return sliverTypes.get(0);
    }

    @Override
    public void setSliverType(SliverType singleType) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getOsImageUrn() {
        if (sliverTypes.size() > 1)
            throw new RuntimeException("The getOsImageUrn convenience method does not support multiple sliver_types");
        if (sliverTypes.isEmpty())
            return null;
        SliverType sliverType = sliverTypes.get(0);
        if (sliverType.getDiskImages().size() > 1)
            throw new RuntimeException("The getOsImageUrn convenience method does not support multiple disk_images in the sliver_type");
        if (sliverType.getDiskImages().isEmpty())
            return null;
        return sliverType.getDiskImages().get(0).getName();
    }

    @Override
    public List<SliverType> getSliverTypes() {
        return sliverTypes;
    }

    @Override
    public NodeLocation getLocation() {
        return nodeLocation;
    }

    @Override
    public void setLocation(NodeLocation value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public GeniUrn getComponentManagerId() {
        return componentManagerId;
    }

    @Override
    public void setComponentManagerId(String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setComponentManagerId(GeniUrn value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public GeniUrn getComponentId() {
        return componentId;
    }

    @Override
    public void setComponentId(String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setComponentId(GeniUrn value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getComponentName() {
        return componentName;
    }

    @Override
    public void setComponentName(String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public GeniUrn getSliverId() {
        return sliverId;
    }

    @Override
    public void setSliverId(GeniUrn sliverId) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setSliverId(String sliverId) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getHostName() {
        return hostname;
    }

    @Override
    public void setHostName(String value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean getRoutableControlIp() {
        return hasRoutableControlIp;
    }

    @Override
    public void setRoutableControlIp(boolean hasRoutableControlIp) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ImmutableExecuteService> getExecuteServices() {
        return executeServices;
    }

    @Override
    public void setExecuteServices(List<? extends ExecuteService> value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ImmutableInstallService> getInstallServices() {
        return installServices;
    }

    @Override
    public void setInstallServices(List<? extends InstallService> value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void addExecuteService(ExecuteService executeService) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void addInstallService(InstallService installService) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<LoginService> getLoginServices() {
        return loginServices;
    }

    @Override
    public void setLoginServices(List<LoginService> value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ProxyService> getProxyServices() {
        return proxyServices;
    }

    @Override
    public void setProxyServices(List<ProxyService> value) {
        throw new UnsupportedOperationException();
    }
    @Override
    public List<WebVncService> getWebVncServices() {
        return webVncServices;
    }

    @Override
    public void setWebVncServices(List<WebVncService> value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ImmutableAnsibleService> getAnsibleServices() {
        return ansibleServices;
    }

    @Override
    public void setAnsibleServices(List<? extends AnsibleService> value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void addAnsibleService(AnsibleService as) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ImmutableRspecInterface> getInterfaces() {
        return interfaces;
    }

    @Override
    public ImmutableRspecInterface getInterfaceByClientId(String clientId) {
        for (ImmutableRspecInterface iface : interfaces) {
            if (Objects.equals(iface.getClientId(), clientId))
                return iface;
        }
        return null;
    }

    @Override
    public ImmutableRspecInterface getInterfaceByUniqueId(String uniqueId) {
        for (ImmutableRspecInterface iface : interfaces) {
            if (Objects.equals(iface.getUniqueId(), uniqueId))
                return iface;
        }
        return null;
    }

    protected boolean addInterface(RspecInterface iface) {
        throw new UnsupportedOperationException();
    }

    protected boolean removeInterface(RspecInterface iface) {
        throw new UnsupportedOperationException();
    }

    @Override
    public List<ImmutableRspecLink> getLinks() {
        return Collections.unmodifiableList(interfaces.stream().filter(ImmutableRspecInterface::isLinkBound)
                .map(ImmutableRspecInterface::getLink)
                .collect(Collectors.toList()));
    }

    @Override
    public List<String> getLeaseIdRefs() {
        return leaseIdRefs;
    }

    @Override
    public void addLeaseIdRef(String leaseIdRef) {
        this.leaseIdRefs.add(leaseIdRef);
    }


    @Override
    public List<String> getAnsibleGroups() {
        return ansibleGroups;
    }

    @Override
    @Nonnull
    public List<ExtraXml> getExtraXml() {
        return extraXmls;
    }

    @Override
    @Nonnull
    public List<ExtraXml> getExtraServicesXml() {
        return extraServicesXmls;
    }


    @Override
    public void addEmptyTag(EmptyTag e) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean hasEmptyTag(EmptyTag e) {
        return emptyTags.contains(e);
    }

    @Override
    public Collection<EmptyTag> getEmptyTags() {
        return emptyTags;
    }

    @Override
    public void overwritePropertiesWith(RspecNode otherNode, boolean overwriteWithNull) {
        throw new UnsupportedOperationException();
    }
}
